import numpy as np
import time
from ccpy.energy.cc_energy import get_cc_energy
from ccpy.lib.core import hbar_cc3, cc3_loops
from ccpy.models.integrals import Integral
from ccpy.models.operators import ClusterOperator

def build_hbar_ccsdta(T, H0, RHF_symmetry, system, *args):
    """Calculate the one- and two-body components of the CCSD(T)(a) similarity-transformed
     Hamiltonian (H_N e^(T1+T2))_C + Delta, where T3 = <ijkabc|(V_N*T2)_C|0>/-D_MP, where
     D_MP = e_a+e_b+e_c-e_i-e_j-e_k."""
    from copy import deepcopy

    save_t3 = True

    nua, noa = T.a.shape
    nub, nob = T.b.shape

    resid_a = np.zeros((nua, noa))
    resid_b = np.zeros((nub, nob))
    resid_aa = np.zeros((nua, nua, noa, noa))
    resid_ab = np.zeros((nua, nub, noa, nob))
    resid_bb = np.zeros((nub, nub, nob, nob))

    # Copy the Bare Hamiltonian object for T1/T2-similarity transformed HBar
    H = deepcopy(H0)

    # Add in the t3-dependent terms to Hbar computed on-the-fly
    H.aa.vooo, H.aa.vvov, H.ab.vooo, H.ab.ovoo, H.ab.vvov, H.ab.vvvo, H.bb.vooo, H.bb.vvov = hbar_cc3.build_hbar(
            H.aa.vooo, H.aa.vvov,
            H.ab.vooo, H.ab.ovoo, H.ab.vvov, H.ab.vvvo,
            H.bb.vooo, H.bb.vvov,
            T.aa, T.ab, T.bb,
            H0.aa.vooo, H0.aa.vvov,
            H0.ab.vooo, H0.ab.ovoo, H0.ab.vvov, H0.ab.vvvo,
            H0.bb.vooo, H0.bb.vvov,
            H0.a.oo, H0.a.vv, H0.b.oo, H0.b.vv,
            H0.aa.oovv, H0.ab.oovv, H0.bb.oovv,
    )

    # It is sometimes useful to store T3 for future use
    if save_t3:
        t_start = time.perf_counter()
        print("   >> Constructing and storing T3[2] = R3 (V_N*T2)_C |0> <<")
        # Prepare MP denominatores
        n = np.newaxis
        eps_o_a = np.diagonal(H0.a.oo); eps_v_a = np.diagonal(H0.a.vv);
        eps_o_b = np.diagonal(H0.b.oo); eps_v_b = np.diagonal(H0.b.vv);
        #
        e3_aaa = (eps_o_a[n, n, n, :, n, n] + eps_o_a[n, n, n, n, :, n] + eps_o_a[n, n, n, n, n, :]
                 -eps_v_a[:, n, n, n, n, n] - eps_v_a[n, :, n ,n, n, n] - eps_v_a[n, n, :, n, n, n])
        e3_aab = (eps_o_a[n, n, n, :, n, n] + eps_o_a[n, n, n, n, :, n] + eps_o_b[n, n, n, n, n, :]
                 -eps_v_a[:, n, n, n, n, n] - eps_v_a[n, :, n ,n, n, n] - eps_v_b[n, n, :, n, n, n])
        e3_abb = (eps_o_a[n, n, n, :, n, n] + eps_o_b[n, n, n, n, :, n] + eps_o_b[n, n, n, n, n, :]
                 -eps_v_a[:, n, n, n, n, n] - eps_v_b[n, :, n ,n, n, n] - eps_v_b[n, n, :, n, n, n])
        e3_bbb = (eps_o_b[n, n, n, :, n, n] + eps_o_b[n, n, n, n, :, n] + eps_o_b[n, n, n, n, n, :]
                 -eps_v_b[:, n, n, n, n, n] - eps_v_b[n, :, n ,n, n, n] - eps_v_b[n, n, :, n, n, n])
        # Prepare new cluster operator
        T_new = ClusterOperator(system, order=3)
        T_new.a = T.a.copy()
        T_new.b = T.b.copy()
        T_new.aa = T.aa.copy()
        T_new.ab = T.ab.copy()
        T_new.bb = T.bb.copy()
        # t(aaa)
        T_new.aaa = -0.25 * np.einsum("amij,bcmk->abcijk", H0.aa.vooo, T.aa, optimize=True)
        T_new.aaa += 0.25 * np.einsum("abie,ecjk->abcijk", H0.aa.vvov, T.aa, optimize=True)
        T_new.aaa /= e3_aaa
        T_new.aaa -= np.transpose(T_new.aaa, (0, 1, 2, 3, 5, 4)) # (jk)
        T_new.aaa -= np.transpose(T_new.aaa, (0, 1, 2, 4, 3, 5)) + np.transpose(T_new.aaa, (0, 1, 2, 5, 4, 3)) # (i/jk)
        T_new.aaa -= np.transpose(T_new.aaa, (0, 2, 1, 3, 4, 5)) # (bc)
        T_new.aaa -= np.transpose(T_new.aaa, (2, 1, 0, 3, 4, 5)) + np.transpose(T_new.aaa, (1, 0, 2, 3, 4, 5)) # (a/bc)
        # t(aab)
        T_new.aab = 0.5 * np.einsum("bcek,aeij->abcijk", H0.ab.vvvo, T.aa, optimize=True)
        T_new.aab -= 0.5 * np.einsum("mcjk,abim->abcijk", H0.ab.ovoo, T.aa, optimize=True)
        T_new.aab += np.einsum("acie,bejk->abcijk", H0.ab.vvov, T.ab, optimize=True)
        T_new.aab -= np.einsum("amik,bcjm->abcijk", H0.ab.vooo, T.ab, optimize=True)
        T_new.aab += 0.5 * np.einsum("abie,ecjk->abcijk", H0.aa.vvov, T.ab, optimize=True)
        T_new.aab -= 0.5 * np.einsum("amij,bcmk->abcijk", H0.aa.vooo, T.ab, optimize=True)
        T_new.aab /= e3_aab
        T_new.aab -= np.transpose(T_new.aab, (1, 0, 2, 3, 4, 5))
        T_new.aab -= np.transpose(T_new.aab, (0, 1, 2, 4, 3, 5))
        # t(abb)
        T_new.abb = 0.5 * np.einsum("abie,ecjk->abcijk", H0.ab.vvov, T.bb, optimize=True)
        T_new.abb -= 0.5 * np.einsum("amij,bcmk->abcijk", H0.ab.vooo, T.bb, optimize=True)
        T_new.abb += 0.5 * np.einsum("cbke,aeij->abcijk", H0.bb.vvov, T.ab, optimize=True)
        T_new.abb -= 0.5 * np.einsum("cmkj,abim->abcijk", H0.bb.vooo, T.ab, optimize=True)
        T_new.abb += np.einsum("abej,ecik->abcijk", H0.ab.vvvo, T.ab, optimize=True)
        T_new.abb -= np.einsum("mbij,acmk->abcijk", H0.ab.ovoo, T.ab, optimize=True)
        T_new.abb /= e3_abb
        T_new.abb -= np.transpose(T_new.abb, (0, 2, 1, 3, 4, 5))
        T_new.abb -= np.transpose(T_new.abb, (0, 1, 2, 3, 5, 4))
        # t(bbb)
        T_new.bbb = -0.25 * np.einsum("amij,bcmk->abcijk", H0.bb.vooo, T.bb, optimize=True)
        T_new.bbb += 0.25 * np.einsum("abie,ecjk->abcijk", H0.bb.vvov, T.bb, optimize=True)
        T_new.bbb /= e3_bbb
        T_new.bbb -= np.transpose(T_new.bbb, (0, 1, 2, 3, 5, 4)) # (jk)
        T_new.bbb -= np.transpose(T_new.bbb, (0, 1, 2, 4, 3, 5)) + np.transpose(T_new.bbb, (0, 1, 2, 5, 4, 3)) # (i/jk)
        T_new.bbb -= np.transpose(T_new.bbb, (0, 2, 1, 3, 4, 5)) # (bc)
        T_new.bbb -= np.transpose(T_new.bbb, (2, 1, 0, 3, 4, 5)) + np.transpose(T_new.bbb, (1, 0, 2, 3, 4, 5)) # (a/bc)
        # replace T
        T = T_new
        t_end = time.perf_counter()
        print(f"   Completed in {t_end - t_start} seconds")

    # Update T1/T2 with T3[2]
    T.a, T.b, T.aa, T.ab, T.bb, _, _, _, _, _ = cc3_loops.update_t(
        T.a, T.b, T.aa, T.ab, T.bb,
        resid_a, resid_b, resid_aa, resid_ab, resid_bb,
        H0.a.oo, H0.a.vv, H0.b.oo, H0.b.vv,
        H0.a.ov, H0.b.ov,
        H0.aa.oovv, H0.ab.oovv, H0.bb.oovv,
        H0.aa.ooov, H0.aa.vovv,
        H0.ab.ooov, H0.ab.oovo, H0.ab.vovv, H0.ab.ovvv,
        H0.bb.ooov, H0.bb.vovv,
        H0.aa.vooo, H0.aa.vvov,
        H0.ab.vooo, H0.ab.ovoo, H0.ab.vvov, H0.ab.vvvo,
        H0.bb.vooo, H0.bb.vvov,
        0.0)

    cc_energy = get_cc_energy(T, H0)
    print(f"   CCSD(T)(a) Correlation Energy: {cc_energy}")
    print(f"   CCSD(T)(a) Total Energy: {cc_energy + system.reference_energy}")

    H.a.ov += (
            np.einsum("imae,em->ia", H0.aa.oovv, T.a, optimize=True)
            + np.einsum("imae,em->ia", H0.ab.oovv, T.b, optimize=True)
    )

    H.a.oo += (
            np.einsum("je,ei->ji", H.a.ov, T.a, optimize=True)
            + np.einsum("jmie,em->ji", H0.aa.ooov, T.a, optimize=True)
            + np.einsum("jmie,em->ji", H0.ab.ooov, T.b, optimize=True)
            + 0.5 * np.einsum("jnef,efin->ji", H0.aa.oovv, T.aa, optimize=True)
            + np.einsum("jnef,efin->ji", H0.ab.oovv, T.ab, optimize=True)
    )

    H.a.vv += (
            - np.einsum("mb,am->ab", H.a.ov, T.a, optimize=True)
            + np.einsum("ambe,em->ab", H0.aa.vovv, T.a, optimize=True)
            + np.einsum("ambe,em->ab", H0.ab.vovv, T.b, optimize=True)
            - 0.5 * np.einsum("mnbf,afmn->ab", H0.aa.oovv, T.aa, optimize=True)
            - np.einsum("mnbf,afmn->ab", H0.ab.oovv, T.ab, optimize=True)
    )

    H.b.ov += (
            np.einsum("imae,em->ia", H0.bb.oovv, T.b, optimize=True)
            + np.einsum("miea,em->ia", H0.ab.oovv, T.a, optimize=True)
    )

    H.b.oo += (
            np.einsum("je,ei->ji", H.b.ov, T.b, optimize=True)
            + np.einsum("jmie,em->ji", H0.bb.ooov, T.b, optimize=True)
            + np.einsum("mjei,em->ji", H0.ab.oovo, T.a, optimize=True)
            + 0.5 * np.einsum("jnef,efin->ji", H0.bb.oovv, T.bb, optimize=True)
            + np.einsum("njfe,feni->ji", H0.ab.oovv, T.ab, optimize=True)
    )

    H.b.vv += (
            - np.einsum("mb,am->ab", H.b.ov, T.b, optimize=True)
            + np.einsum("ambe,em->ab", H0.bb.vovv, T.b, optimize=True)
            + np.einsum("maeb,em->ab", H0.ab.ovvv, T.a, optimize=True)
            - 0.5 * np.einsum("mnbf,afmn->ab", H0.bb.oovv, T.bb, optimize=True)
            - np.einsum("nmfb,fanm->ab", H0.ab.oovv, T.ab, optimize=True)
    )

    Q1 = -np.einsum("mnfe,an->amef", H0.aa.oovv, T.a, optimize=True)
    I2A_vovv = H0.aa.vovv + 0.5 * Q1
    H.aa.vovv = I2A_vovv + 0.5 * Q1

    Q1 = np.einsum("mnfe,fi->mnie", H0.aa.oovv, T.a, optimize=True)
    I2A_ooov = H0.aa.ooov + 0.5 * Q1
    H.aa.ooov = I2A_ooov + 0.5 * Q1

    Q1 = -np.einsum("nmef,an->amef", H0.ab.oovv, T.a, optimize=True)
    I2B_vovv = H0.ab.vovv + 0.5 * Q1
    H.ab.vovv = I2B_vovv + 0.5 * Q1

    Q1 = np.einsum("mnfe,fi->mnie", H0.ab.oovv, T.a, optimize=True)
    I2B_ooov = H0.ab.ooov + 0.5 * Q1
    H.ab.ooov = I2B_ooov + 0.5 * Q1

    Q1 = -np.einsum("mnef,an->maef", H0.ab.oovv, T.b, optimize=True)
    I2B_ovvv = H0.ab.ovvv + 0.5 * Q1
    H.ab.ovvv = I2B_ovvv + 0.5 * Q1

    Q1 = np.einsum("nmef,fi->nmei", H0.ab.oovv, T.b, optimize=True)
    I2B_oovo = H0.ab.oovo + 0.5 * Q1
    H.ab.oovo = I2B_oovo + 0.5 * Q1

    Q1 = -np.einsum("nmef,an->amef", H0.bb.oovv, T.b, optimize=True)
    I2C_vovv = H0.bb.vovv + 0.5 * Q1
    H.bb.vovv = I2C_vovv + 0.5 * Q1

    Q1 = np.einsum("mnfe,fi->mnie", H0.bb.oovv, T.b, optimize=True)
    I2C_ooov = H0.bb.ooov + 0.5 * Q1
    H.bb.ooov = I2C_ooov + 0.5 * Q1

    Q1 = -np.einsum("bmfe,am->abef", I2A_vovv, T.a, optimize=True)
    Q1 -= np.transpose(Q1, (1, 0, 2, 3))
    H.aa.vvvv += 0.5 * np.einsum("mnef,abmn->abef", H0.aa.oovv, T.aa, optimize=True) + Q1

    H.ab.vvvv += (
            - np.einsum("mbef,am->abef", I2B_ovvv, T.a, optimize=True)
            - np.einsum("amef,bm->abef", I2B_vovv, T.b, optimize=True)
            + np.einsum("mnef,abmn->abef", H0.ab.oovv, T.ab, optimize=True)
    )

    Q1 = -np.einsum("bmfe,am->abef", I2C_vovv, T.b, optimize=True)
    Q1 -= np.transpose(Q1, (1, 0, 2, 3))
    H.bb.vvvv += 0.5 * np.einsum("mnef,abmn->abef", H0.bb.oovv, T.bb, optimize=True) + Q1

    Q1 = +np.einsum("nmje,ei->mnij", I2A_ooov, T.a, optimize=True)
    Q1 -= np.transpose(Q1, (0, 1, 3, 2))
    H.aa.oooo += 0.5 * np.einsum("mnef,efij->mnij", H0.aa.oovv, T.aa, optimize=True) + Q1

    H.ab.oooo += (
            np.einsum("mnej,ei->mnij", I2B_oovo, T.a, optimize=True)
            + np.einsum("mnie,ej->mnij", I2B_ooov, T.b, optimize=True)
            + np.einsum("mnef,efij->mnij", H0.ab.oovv, T.ab, optimize=True)
    )

    Q1 = +np.einsum("nmje,ei->mnij", I2C_ooov, T.b, optimize=True)
    Q1 -= np.transpose(Q1, (0, 1, 3, 2))
    H.bb.oooo += 0.5 * np.einsum("mnef,efij->mnij", H0.bb.oovv, T.bb, optimize=True) + Q1

    H.aa.voov += (
            np.einsum("amfe,fi->amie", I2A_vovv, T.a, optimize=True)
            - np.einsum("nmie,an->amie", I2A_ooov, T.a, optimize=True)
            + np.einsum("nmfe,afin->amie", H0.aa.oovv, T.aa, optimize=True)
            + np.einsum("mnef,afin->amie", H0.ab.oovv, T.ab, optimize=True)
    )

    H.ab.voov += (
            np.einsum("amfe,fi->amie", I2B_vovv, T.a, optimize=True)
            - np.einsum("nmie,an->amie", I2B_ooov, T.a, optimize=True)
            + np.einsum("nmfe,afin->amie", H0.ab.oovv, T.aa, optimize=True)
            + np.einsum("nmfe,afin->amie", H0.bb.oovv, T.ab, optimize=True)
    )

    H.ab.ovvo += (
            np.einsum("maef,fi->maei", I2B_ovvv, T.b, optimize=True)
            - np.einsum("mnei,an->maei", I2B_oovo, T.b, optimize=True)
            + np.einsum("mnef,afin->maei", H0.ab.oovv, T.bb, optimize=True)
            + np.einsum("mnef,fani->maei", H0.aa.oovv, T.ab, optimize=True)
    )

    H.ab.ovov += (
            np.einsum("mafe,fi->maie", I2B_ovvv, T.a, optimize=True)
            - np.einsum("mnie,an->maie", I2B_ooov, T.b, optimize=True)
            - np.einsum("mnfe,fain->maie", H0.ab.oovv, T.ab, optimize=True)
    )

    H.ab.vovo += (
            - np.einsum("nmei,an->amei", I2B_oovo, T.a, optimize=True)
            + np.einsum("amef,fi->amei", I2B_vovv, T.b, optimize=True)
            - np.einsum("nmef,afni->amei", H0.ab.oovv, T.ab, optimize=True)
    )

    H.bb.voov += (
            np.einsum("amfe,fi->amie", I2C_vovv, T.b, optimize=True)
            - np.einsum("nmie,an->amie", I2C_ooov, T.b, optimize=True)
            + np.einsum("nmfe,afin->amie", H0.bb.oovv, T.bb, optimize=True)
            + np.einsum("nmfe,fani->amie", H0.ab.oovv, T.ab, optimize=True)
    )

    Q1 = (
            np.einsum("mnjf,afin->amij", H.aa.ooov, T.aa, optimize=True)
            + np.einsum("mnjf,afin->amij", H.ab.ooov, T.ab, optimize=True)
    )
    Q2 = H0.aa.voov + 0.5 * np.einsum("amef,ei->amif", H0.aa.vovv, T.a, optimize=True)
    Q2 = np.einsum("amif,fj->amij", Q2, T.a, optimize=True)
    Q1 += Q2
    Q1 -= np.transpose(Q1, (0, 1, 3, 2))
    H.aa.vooo += Q1 + (
            np.einsum("me,aeij->amij", H.a.ov, T.aa, optimize=True)
            - np.einsum("nmij,an->amij", H.aa.oooo, T.a, optimize=True)
            + 0.5 * np.einsum("amef,efij->amij", H0.aa.vovv, T.aa, optimize=True)
    )

    Q1 = H0.ab.voov + np.einsum("amfe,fi->amie", H0.ab.vovv, T.a, optimize=True)
    H.ab.vooo += (
            np.einsum("me,aeij->amij", H.b.ov, T.ab, optimize=True)
            - np.einsum("nmij,an->amij", H.ab.oooo, T.a, optimize=True)
            + np.einsum("mnjf,afin->amij", H.bb.ooov, T.ab, optimize=True)
            + np.einsum("nmfj,afin->amij", H.ab.oovo, T.aa, optimize=True)
            - np.einsum("nmif,afnj->amij", H.ab.ooov, T.ab, optimize=True)
            + np.einsum("amej,ei->amij", H0.ab.vovo, T.a, optimize=True)
            + np.einsum("amie,ej->amij", Q1, T.b, optimize=True)
            + np.einsum("amef,efij->amij", H0.ab.vovv, T.ab, optimize=True)
    )

    Q1 = H0.ab.ovov + np.einsum("mafe,fj->maje", H0.ab.ovvv, T.a, optimize=True)
    H.ab.ovoo += (
            np.einsum("me,eaji->maji", H.a.ov, T.ab, optimize=True)
            - np.einsum("mnji,an->maji", H.ab.oooo, T.b, optimize=True)
            + np.einsum("mnjf,fani->maji", H.aa.ooov, T.ab, optimize=True)
            + np.einsum("mnjf,fani->maji", H.ab.ooov, T.bb, optimize=True)
            - np.einsum("mnfi,fajn->maji", H.ab.oovo, T.ab, optimize=True)
            + np.einsum("maje,ei->maji", Q1, T.b, optimize=True)
            + np.einsum("maei,ej->maji", H0.ab.ovvo, T.a, optimize=True)
            + np.einsum("mafe,feji->maji", H0.ab.ovvv, T.ab, optimize=True)
    )

    Q1 = (
            np.einsum("mnjf,afin->amij", H.bb.ooov, T.bb, optimize=True)
            + np.einsum("nmfj,fani->amij", H.ab.oovo, T.ab, optimize=True)
    )
    Q2 = H0.bb.voov + 0.5 * np.einsum("amef,ei->amif", H0.bb.vovv, T.b, optimize=True)
    Q2 = np.einsum("amif,fj->amij", Q2, T.b, optimize=True)
    Q1 += Q2
    Q1 -= np.transpose(Q1, (0, 1, 3, 2))
    H.bb.vooo += Q1 + (
            + np.einsum("me,aeij->amij", H.b.ov, T.bb, optimize=True)
            - np.einsum("nmij,an->amij", H.bb.oooo, T.b, optimize=True)
            + 0.5 * np.einsum("amef,efij->amij", H0.bb.vovv, T.bb, optimize=True)
    )

    Q1 = (
            np.einsum("bnef,afin->abie", H.aa.vovv, T.aa, optimize=True)
            + np.einsum("bnef,afin->abie", H.ab.vovv, T.ab, optimize=True)
    )
    Q2 = H0.aa.ovov - 0.5 * np.einsum("mnie,bn->mbie", H0.aa.ooov, T.a, optimize=True)
    Q2 = -np.einsum("mbie,am->abie", Q2, T.a, optimize=True)
    Q1 += Q2
    Q1 -= np.transpose(Q1, (1, 0, 2, 3))
    H.aa.vvov += Q1 + (
            - np.einsum("me,abim->abie", H.a.ov, T.aa, optimize=True)
            + np.einsum("abfe,fi->abie", H.aa.vvvv, T.a, optimize=True)
            + 0.5 * np.einsum("mnie,abmn->abie", H0.aa.ooov, T.aa, optimize=True)
    )

    Q1 = H0.ab.ovov - np.einsum("mnie,bn->mbie", H0.ab.ooov, T.b, optimize=True)
    Q1 = -np.einsum("mbie,am->abie", Q1, T.a, optimize=True)
    H.ab.vvov += Q1 + (
            - np.einsum("me,abim->abie", H.b.ov, T.ab, optimize=True)
            + np.einsum("abfe,fi->abie", H.ab.vvvv, T.a, optimize=True)
            + np.einsum("nbfe,afin->abie", H.ab.ovvv, T.aa, optimize=True)
            + np.einsum("bnef,afin->abie", H.bb.vovv, T.ab, optimize=True)
            - np.einsum("amfe,fbim->abie", H.ab.vovv, T.ab, optimize=True)
            - np.einsum("amie,bm->abie", H0.ab.voov, T.b, optimize=True)
            + np.einsum("nmie,abnm->abie", H0.ab.ooov, T.ab, optimize=True)
    )

    Q1 = H0.ab.vovo - np.einsum("nmei,bn->bmei", H0.ab.oovo, T.a, optimize=True)
    Q1 = -np.einsum("bmei,am->baei", Q1, T.b, optimize=True)
    H.ab.vvvo += Q1 + (
            - np.einsum("me,bami->baei", H.a.ov, T.ab, optimize=True)
            + np.einsum("baef,fi->baei", H.ab.vvvv, T.b, optimize=True)
            + np.einsum("bnef,fani->baei", H.aa.vovv, T.ab, optimize=True)
            + np.einsum("bnef,fani->baei", H.ab.vovv, T.bb, optimize=True)
            - np.einsum("maef,bfmi->baei", H.ab.ovvv, T.ab, optimize=True)
            - np.einsum("naei,bn->baei", H0.ab.ovvo, T.a, optimize=True)
            + np.einsum("nmei,banm->baei", H0.ab.oovo, T.ab, optimize=True)
    )

    Q1 = (
            np.einsum("bnef,afin->abie", H.bb.vovv, T.bb, optimize=True)
            + np.einsum("nbfe,fani->abie", H.ab.ovvv, T.ab, optimize=True)
    )
    Q2 = H.bb.ovov - 0.5 * np.einsum("mnie,bn->mbie", H0.bb.ooov, T.b, optimize=True)
    Q2 = -np.einsum("mbie,am->abie", Q2, T.b, optimize=True)
    Q1 += Q2
    Q1 -= np.transpose(Q1, (1, 0, 2, 3))
    H.bb.vvov += Q1 + (
            - np.einsum("me,abim->abie", H.b.ov, T.bb, optimize=True)
            + np.einsum("abfe,fi->abie", H.bb.vvvv, T.b, optimize=True)
            + 0.5 * np.einsum("mnie,abmn->abie", H0.bb.ooov, T.bb, optimize=True)
    )

    # For RHF symmetry, copy a parts to b and aa parts to bb
    if RHF_symmetry:
        H.b.ov = H.a.ov.copy()
        H.b.oo = H.a.oo.copy()
        H.b.vv = H.a.vv.copy()
        H.bb.oooo = H.aa.oooo.copy()
        H.bb.ooov = H.aa.ooov.copy()
        H.bb.vooo = H.aa.vooo.copy()
        H.bb.oovv = H.aa.oovv.copy()
        H.bb.voov = H.aa.voov.copy()
        H.bb.vovv = H.aa.vovv.copy()
        H.bb.vvov = H.aa.vvov.copy()
        H.bb.vvvv = H.aa.vvvv.copy()

    return H, T, cc_energy
